#coding: UTF-8

from wb import DefineModule
import grt
import os
import tempfile
import time

from mforms import Utilities, ResultOk, AppView, newSectionBox, App
import mforms

import wb_admin_monitor
import wb_admin_configuration
import wb_admin_utils
import wb_admin_control_be
from wb_admin_utils import MySQLConnection, MySQLError

# define this Python module as a GRT module
ModuleInfo = DefineModule(name= "WbAdmin", author= "Sun Microsystems Inc.", version="1.0")

class DBError(Exception):
    pass

tab_references = []




##def get_db_connection(server_instance_settings):
#  if server_instance_settings.connection:
#    db_connection = MySQLConnection(server_instance_settings.connection)
#    ignore_error = False
#    error_location = None
#    the_error = None
#    try:
#      db_connection.connect()
#    except MySQLError, exc:
#     # errors that probably just mean the server is down can be ignored (ex 2013)
#     # errors from incorrect connect parameters should raise an exception
#     # ex 1045: bad password
#      if exc.code in (1045,):
#        raise exc
#      elif exc.code in (2013,):
#        ignore_error = True
#      error_location = exc.location
#      the_error = str(exc)
#
#      if not ignore_error:
#        if Utilities.show_warning("Could not Connect to MySQL Server at %s" % error_location,
#                "%s\nYou can continue but some functionality may be unavailable." % the_error,
#                "Continue Anyway", "Cancel", "") != ResultOk:
#          raise MySQLError("")
#    return db_connection
#  else:
#    Utilities.show_warning("WB Admin", "Server Profile has no database connection specified.\nSome functionality will not be available.", "OK", "", "")
#
#
#  return None


class AdministratorTab(AppView):
    #top_box holds all sections
    top_vbox = None
    monitor = None
    configuration = None
    ctrl_be = None
    
    def __init__(self, server_instance_settings):
        AppView.__init__(self, False, "admin", True)
        self.ctrl_be = wb_admin_control_be.MyCtrl(server_instance_settings)

        self.on_close(wb_admin_utils.weakcb(self, "handle_on_close"))

        # Create sections and add them to the admin page.
        self.monitor = wb_admin_monitor.WbAdminMonitor(server_instance_settings, self.ctrl_be)
        section = self.add_section(u"サーバの状況", True)
        section.set_content(self.monitor)

        self.configuration = wb_admin_configuration.WbAdminConfiguration(server_instance_settings, self.ctrl_be, self.monitor)
        section = self.add_section("設定", False)
        section.set_content(self.configuration)


    def handle_on_close(self):
        App.get().set_status_text(u"閉じています。")
        self.configuration.shutdown()
        tab_references.remove(self)
        return True


    def wait_server_check(self, timeout):
        # wait for the server status check to be performed until the given timeout
        t = time.time()
        while self.configuration.last_is_running_check is None and time.time() - t < timeout:
            time.sleep(0.5)

    
    def add_section(self, label_text, is_content_expandable):
        section = newSectionBox(is_content_expandable, label_text)
        section.set_name(label_text)
        self.add(section, not is_content_expandable, True)
        return section



def do_open_administrator(server_instance):
    app = App.get()
    try:
      adminTab = AdministratorTab(server_instance)
    except wb_admin_control_be.ConnectionError, exc:
      Utilities.show_error(u"サーバ接続中にエラーが発生しました (%s@%s)" % (server_instance.loginInfo["ssh.userName"], server_instance.loginInfo["ssh.hostName"]), str(exc), "OK", "", "")
      app.set_status_text(u"WBアドミンを開くことができません")
      return 0
    except wb_admin_utils.MySQLError, exc:
      if exc.message:
        Utilities.show_error(u"MySQLサーバ (%s) に接続中にエラーが発生しました" % exc.location, str(exc), "OK", "", "")
      app.set_status_text(u"WBアドミンを開くことができません")
      return 0
    
    app.dock_view(adminTab, "maintab")
    app.set_view_title(adminTab, u"アドミン (%s)" % (server_instance.name))

    tab_references.append(adminTab)

    app.set_status_text(u"WBアドミンを開きました")

    return adminTab


@ModuleInfo.plugin("wb.admin.open", type="standalone", caption= u"WBアドミンを初期化",  pluginMenu= "Administrator")
@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_ServerInstance)
def openAdministrator(server_instance):
    do_open_administrator(server_instance)
    return 1


@ModuleInfo.plugin("wb.admin.filterDebugger", type="standalone", caption= u"テストフィルタ",  pluginMenu= "Utilities")
@ModuleInfo.export(grt.INT)
def openFilterDebugger():
    wb_admin_control_be.run_filter_debugger()
    return 1


def selectServer(title):
    window = mforms.Form(None)
    window.set_title(title)
    box = mforms.newBox(False)
    window.set_content(box)
    
    box.set_padding(12)
    box.set_spacing(12)
    
    label = mforms.newLabel()
    label.set_text(u"接続するサーバを選択してください：")
    box.add(label, False, True)
    
    listbox = mforms.newListBox(False)
    box.add(listbox, True, True)

    for inst in grt.root.wb.rdbmsMgmt.storedInstances:
      listbox.add_item(inst.name)
    
    bbox = mforms.newBox(True)
    box.add(bbox, False, True)
    
    bbox.set_spacing(8)
    
    ok = mforms.newButton()
    ok.set_text("OK")
    bbox.add_end(ok, False, True)

    cancel = mforms.newButton()
    cancel.set_text(u"ｷｬﾝｾﾙ")
    bbox.add_end(cancel, False, True)
    
    window.set_size(300, 300)
    window.center()
    
    if window.run_modal(ok, cancel):
      i = listbox.get_selected_index()
      if i >= 0:
          return grt.root.wb.rdbmsMgmt.storedInstances[i]
    return None
        


@ModuleInfo.plugin("wb.admin.dumpManager", type="standalone", caption= u"データベースのエキスポート／インポートマネージャを開く",  pluginMenu= "Administrator")
@ModuleInfo.export(grt.INT)
def openExportImport():
    server = selectServer("Import/Export MySQL Data")
    if server:
        tab = do_open_administrator(server)        
        tab.wait_server_check(4)
        tab.configuration.switch_to(5)
    return 1




@ModuleInfo.plugin("wb.admin.securityManager", type="standalone", caption= u"セキュリティマネージャを開く",  pluginMenu= "Administrator")
@ModuleInfo.export(grt.INT)
def openSecurityManager():
    server = selectServer("Security Manager")
    if server:
        tab = do_open_administrator(server)
        tab.wait_server_check(4)
        tab.configuration.switch_to(2)
    return 1


@ModuleInfo.plugin("wb.admin.placeholder", type="standalone", caption= u"アドミンを開く",  pluginMenu= "Administrator")
@ModuleInfo.export(grt.INT)
def openSecurityManager():
    server = selectServer("Administrator")
    if server:
        tab = do_open_administrator(server)
    return 1
  

def check_if_config_file_has_section(config_file, section):
    for line in config_file:
        if line.strip() == "[%s]"%section:
	    return True
    return False


test_ssh_connection = None
test_ssh_connection_is_windows = None

@ModuleInfo.export(grt.STRING, grt.STRING, grt.classes.db_mgmt_ServerInstance, grt.STRING)
def testInstanceSettingByName(what, server_instance, ssh_password):
    global test_ssh_connection
    from wb_admin_ssh import WbAdminSSH
  
    if what == "connect_to_host":
        if test_ssh_connection:
            test_ssh_connection.close()

        loginInfo = server_instance.loginInfo

        test_ssh_connection = WbAdminSSH()

        try:
            grt.send_info(u"%s に接続中" % loginInfo['ssh.hostName'])
            test_ssh_connection.connect(loginInfo['ssh.hostName'], loginInfo['ssh.port'], loginInfo['ssh.userName'], loginInfo['ssh.password'] or ssh_password, loginInfo['ssh.useKey'], loginInfo['ssh.key'])
            grt.send_info("connected.")
        except Exception, exc:
          import traceback
          traceback.print_exc()
          return "ERROR "+str(exc)
        except:
          print u"不明なエラー"
          return "ERROR"

        if not test_ssh_connection.is_connected():
            test_ssh_connection = None
            return u"ホストに接続中にエラーが発生しました。"

        result, rc = test_ssh_connection.exec_cmd("uname")
        if rc != 0:
            # error calling uname, possibly windows
            result, rc = test_ssh_connection.exec_cmd("ver")
            test_ssh_connection_is_windows = True
        else:
            test_ssh_connection_is_windows = False
        print "ＯＫ。OSは'%s'です" % result.strip()
        if not result:
            return "ERROR"

        return "OK " + result.strip()

    elif what == "disconnect":
        if test_ssh_connection:
            test_ssh_connection.close()
            test_ssh_connection = None
        return "OK"
    
    elif what == "check_privileges":
        return "ERROR"
    
    elif what == "find_config_file" or what == "check_ini_path":
        print "Check if %s exists" % server_instance.serverInfo["sys.config.path"]
        if test_ssh_connection.file_exists(server_instance.serverInfo["sys.config.path"]):
            return "OK"
        return "ERROR File %s doesn't exist" % server_instance.serverInfo["sys.config.path"]

    elif what == "find_config_file/local" or what == "check_ini_path/local":
        print "Check if %s exists locally" % server_instance.serverInfo["sys.config.path"]
        if os.path.exists(server_instance.serverInfo["sys.config.path"]):
            return "OK"
        return "ERROR File %s doesn't exist" % server_instance.serverInfo["sys.config.path"]

    elif what == "find_error_files":
        return "ERROR"

    elif what == "check_ini_section/local":
      config_file= server_instance.serverInfo["sys.config.path"]
      print "Check if %s exists locally" % config_file
      if not os.path.exists(config_file):
        return "ERROR File %s doesn't exist" % config_file

      section = server_instance.serverInfo["sys.config.section"]
      print "Check if %s section exists in %s" % (section, config_file)
      #parser = ConfigParser.ConfigParser()
      #parser.read([config_file])
      #if parser.has_section(section):
      #  return "OK"
      # can't use ConfigParser because this pos doesn't like options without an assignment
      if check_if_config_file_has_section(open(config_file, "r"), section):
	return "OK"
      return "ERROR Couldn't find section %s in the config file %s" % (section, config_file)
      
    elif what == "check_ini_section":
      config_file = server_instance.serverInfo["sys.config.path"]
      print "Check if %s exists" % config_file
      if not test_ssh_connection.file_exists(config_file):
        return "ERROR File %s doesn't exist" % config_file
      
      section = server_instance.serverInfo["sys.config.section"]
      print "Check if %s section exists in %s" % (section, config_file)
      local_file = tempfile.NamedTemporaryFile()
      test_ssh_connection.get(config_file, local_file.name)
      #parser = ConfigParser.ConfigParser()
      #parser.readfp(local_file)
      #if parser.has_section(section):
      #  return "OK"
      if check_if_config_file_has_section(local_file, section):
        return "OK"
      return "ERROR Couldn't find section %s in the config file %s" % (section, config_file)
      
    elif what == "check_admin_commands":
        path = server_instance.serverInfo["sys.mysqld.start"]
        cmd_start= None
        if path.startswith("/"):
            cmd_start = path.split()[0]
            if not test_ssh_connection.file_exists(cmd_start):
                return "ERROR %s is invalid" % path

        path = server_instance.serverInfo["sys.mysqld.stop"]
        if path.startswith("/"):
            cmd = path.split()[0]
            if cmd != cmd_start and not test_ssh_connection.file_exists(cmd):
                return "ERROR %s is invalid" % path

        command = server_instance.serverInfo["sys.mysqld.status"]
        print "Checking command '%s'" % command
        result, rc = test_ssh_connection.exec_cmd(command)
        print "Server detected as %s" % (rc and "running" or "stopped")

        return "OK"

    elif what == "check_admin_commands/local":
        path = server_instance.serverInfo["sys.mysqld.start"]
        cmd_start= None
        if path.startswith("/"):
            cmd_start = path.split()[0]
            if not os.path.exists(cmd_start):
                return "ERROR %s is invalid" % path

        path = server_instance.serverInfo["sys.mysqld.stop"]
        if path.startswith("/"):
            cmd = path.split()[0]
            if cmd != cmd_start and not os.path.exists(cmd):
                return "ERROR %s is invalid" % path

        command = server_instance.serverInfo["sys.mysqld.status"]
        print "Checking command '%s'" % command
        result, rc = wb_admin_control_be.run_cmd(command, 0, None)
        print "Server detected as %s" % (rc and "running" or "stopped")

        return "OK"

    return "ERROR"


@ModuleInfo.export(grt.DICT, grt.classes.db_mgmt_ServerInstance)
def detectInstanceSettings(server_instance):    
    #form = Form()
    
    #form.run(None, None)
    
    return {}



@ModuleInfo.export(grt.INT, grt.classes.db_mgmt_ServerInstance)
def testInstanceSettings(server_instance):
    
    return 0

